import math
from typing import List, Union
import numpy as np
from warnings import warn
from . import models
from PIL import Image, ImageDraw
from shapely.geometry import Polygon

def deg2num(lat_deg: float, lon_deg:  float, zoom: int):
    """Convert lat,lng to xyz tile coordinates.
    reference: https://developers.planet.com/docs/planetschool/xyz-tiles-and-slippy-maps/

    Args:
        lat_deg (float): latitude in decimal degrees
        lon_deg (float): longitude in decimal degrees
        zoom (int): zoom level e.g. 19

    Returns:
        x_tile (float): x in tile coordinates at specified zoom level
        y_tile (float): y in tile coordinates at specified zoom level
    """

    lat_rad = math.radians(lat_deg)
    n = 2.0 ** zoom
    xtile = (lon_deg + 180.0) / 360.0 * n
    ytile = (1.0 - math.asinh(math.tan(lat_rad)) / math.pi) / 2.0 * n
    return xtile, ytile



def num2deg(xtile: Union[float, int], ytile: Union[float, int], zoom: int):
    """convert x, y, tile coordinates to lat, lng. 
    ref: https://developers.planet.com/docs/planetschool/xyz-tiles-and-slippy-maps/

    Args:
        xtile (Union[float, int]): x position
        ytile (Union[float, int]): y position
        zoom (int): zoom level e.g. 19

    Returns:
        lat_deg (float): latitude in decimal degrees
        lng_deg (float): longitude in decimal degrees
    """
    n = 2.0 ** zoom
    lng_deg = xtile / n * 360.0 - 180.0
    lat_rad = math.atan(math.sinh(math.pi * (1 - 2 * ytile / n)))
    lat_deg = math.degrees(lat_rad)
    return lat_deg, lng_deg



def getBboxLatLng(coords: List[List]):
    """Get bounding box for a polygon.

    Args:
        coords (List[List]): list of points (x,y) = (lng,lat) with bottom left reference (e.g. [[lng,lat], [lng,lat]])

    Returns:
        x1 (float): closest distance from left reference
        y1 (float): closest distance from top reference
        x2 (float): furthest distance from left reference
        y2 (float): furthest distance from top reference

    """    

    coords = np.array(coords)
    x1 = np.min(coords[:,0])
    y1 = np.max(coords[:,1])
    x2 = np.max(coords[:,0])
    y2 = np.min(coords[:,1])

    return x1, y1, x2, y2



def getBboxTileCoords(coords: List[List], zxy: str):
    """Get bounding box for a polygon. output is in tile coordinates

    Args:
        coords (List[List]): list of points (x,y) = (lng,lat) with bottom left reference (e.g. [[lng,lat], [lng,lat]])

    Returns:
        x1 (float): closest distance from tile left side
        y1 (float): closest distance from tile top side
        x2 (float): furthest distance from tile left side
        y2 (float): furthest distance from tile top side

    """    

    # get bounding box (top left reference)
    x1, y1, x2, y2 = getBboxLatLng(coords)

    z = int(zxy.split("/")[0])
    x = int(zxy.split("/")[1])
    y = int(zxy.split("/")[2])

    # get annotation bounding box in tile coordinates
    x1, y1 = deg2num(y1, x1, z)
    x2, y2 = deg2num(y2, x2, z)

    x1 -= x
    x2 -= x
    y1 -= y
    y2 -= y

    return x1, y1, x2, y2


def getOverlap(coords: List[List], zxy: str):
    """ Get the portion of a bounding box that is inside a tile

    Args:
        coords (List[List]): list of points in the polygon (lat,lng)
        zxy (str): tile e.g. 12/345/678

    Returns:
        float: overlap [0,1]
    """
    x1, y1, x2, y2 = getBboxTileCoords(coords, zxy)

    # get intersection area
    x_overlap = np.clip(x2, 0, 1) - np.clip(x1, 0, 1)
    y_overlap = np.clip(y2, 0, 1) - np.clip(y1, 0, 1)

    # get annotation area
    annotation_area = abs((x2-x1)*(y2-y1))
    
    if annotation_area == 0:
        return 0

    return (x_overlap*y_overlap) / annotation_area


def getAnnotationsForTile(
            annotations: List[models.Annotation],
            zxy: str,
            overlap_threshold: float = 0.2
        ) -> List[models.Annotation]:
        """ Filter a set of annotations for those that have overlap with some tile

        Args:
            annotations (List[Annotation]): Annotations to filter
            zxy (str): The tile e.g. 12/345/678
            overlap_threshold (float, optional): How much overlap. Defaults to 0.2.

        Returns:
            List[Annotation]: All the annotations that have enough overlap with the specified tile
        """        

        annotationsInTile = []

        # filter annotations
        for annotation in annotations:
            # check overlap with tile
            overlap = getOverlap(annotation.coordinates, zxy)
            if overlap < overlap_threshold:
                continue
            annotationsInTile.append(annotation)

        return annotationsInTile

def bboxFromCoords(coordinates: List[List], zxy: str, tile_size: int, clip: bool = False):
    """ Get a bounding box from a polygon (lng, lat)

    Args:
        coordinates (List[List]): List of points in polygon (lng, lat)
        zxy (str): Tile zxy
        tile_size (int): size of tile, e.g. 256
        clip (Optional[bool]): Clip the boxes to the edge of the tiles

    Returns:
        x1 (int): top-left x coordinate
        y1 (int): top-left y coordinate
        x2 (int): bottom-right x coordinate
        y2 (int): bottom-right y coordinate
    """

    bbox = getBboxTileCoords(coordinates, zxy)
    
    [x1, y1, x2, y2] = bbox
    
    if clip:
        x1 = np.clip(bbox[0], 0, 1)
        y1 = np.clip(bbox[1], 0, 1)
        x2 = np.clip(bbox[2], 0, 1)
        y2 = np.clip(bbox[3], 0, 1)

    x1 *= tile_size
    y1 *= tile_size
    x2 *= tile_size
    y2 *= tile_size

    return x1, y1, x2, y2


def bboxToCoco(x1: int, y1: int, x2: int, y2: int):
    """ Convert two-point bbox to coco format

    Args:
        x1 (int): top-left x coordinate
        y1 (int): top-left y coordinate
        x2 (int): bottom-right x coordinate
        y2 (int): bottom-right y coordinate

    Returns:
        x (int): top-left x coordinate
        y (int): top-left y coordinate
        w (int): box width
        h (int): box height
    """

    w = int(x2-x1)
    h = int(y2-y1)
    
    return x1, y1, w, h



def yx_to_xy(yx: List[List]) -> List[List]:
    """" switch the ordering in our array, explicitely so we dont forget

    Args:
        yx (List[List]): [[y,x], [y,x]]

    Returns:
        List[List]: [[x,y], [x,y]]
    """
    return [[x,y] for y,x in yx]

def bboxToPolygon(x1, y1, x2, y2):
    """ Get a 4 point polygon from a bounding box.

    Args:
        x1 (float): left edge
        y1 (float): top edge
        x2 (float): right edge
        y2 (float): bottom edge

    Returns:
        List[List[float]]: polygon
    """
    return [[x1, y1], [x2, y1], [x2, y2], [x1, y2], [x1, y1]]

def coordsFromPolygon(
        polygon: List[List[float]],
        zxy: str, 
        tile_size: int,
        padding: int = 0) -> List[List[float]]:
    """ get a lat/lng polygon from a polygon in a tile

    Args:
        polygon (List[List[float]]): coordinates in the image [[y,x],[y,x]]
        zxy (str): zxy string for the tile e.g. 12/34/567
        width (int): width of the tile
        height (int): height of the tile
        padding (int, optional): number of extra pixels around the edge of the tile. Defaults to 0.

    Returns:
        List[List[float]]: coordinates in [[lng,lat],[lng,lat]] format
    """


    z,x,y = splitZXY(zxy)

    points_lat_lng = []
    for (x1,y1) in polygon:

      x1 -= padding
      y1 -= padding

      # Normalise
      x1 /= tile_size
      y1 /= tile_size

      # add task corner x,y
      x1 += x
      y1 += y

      # convert to lat,lng
      lat, lng = num2deg(x1, y1, z)
      points_lat_lng.append([lng, lat])

    return points_lat_lng

def coordsFromBbox(
        x1: int, 
        y1: int, 
        x2: int, 
        y2: int, 
        zxy: str, 
        width: int, 
        height: int) -> List[List[float]]:
    """Get lng,lat coordinate list from a bounding box and width height

    Args:
        x1 (int): left edge
        y1 (int): top edge
        x2 (int): right edge
        y2 (int): bottom edge
        zxy (str): tile e.g. 12/345/678
        width (int): tile width in pixels
        height (int): tile height in pixels


    Returns:
        List[List[float]]: coordinate for polygon

    """

    warn('This function is deprecated, consider a combination of bboxToPolygon and coordsFromPolygon', DeprecationWarning, stacklevel=2)


    z,x,y = splitZXY(zxy)
    x1 /= width
    y1 /= height
    x2 /= width
    y2 /= height
    points = [[x+x1, y+y1],
        [x+x2, y+y1],
        [x+x2, y+y2],
        [x+x1, y+y2],
        [x+x1, y+y1]]
    points_lat_lng = []
    for point in points:
        lat, lng = num2deg(point[0], point[1], z)
        points_lat_lng.append([lng, lat])
    return points_lat_lng


def latLngToImgCoords(coords: List[List], zxy: str, tile_size: int) -> List[List]:
    """ convert a list of coordinates to pixel coordinates in a tile

    Args:
        coords (List[List]): coordinates e.g. [[lng,lat], [lng, lat]]
        zxy (str): tile zxy string e.g. 12/34/567
        tile_size (int): width of the tile e.g. 256 - dont forget padding

    Returns:
        List[List]: _description_
    """

    tile_z, tile_x, tile_y = splitZXY(zxy)

    coords_tile = []
    # get annotation bounding box in tile coordinates
    for (x, y) in coords:
      x_prime, y_prime = deg2num(y, x, tile_z)
      coords_tile.append([x_prime - tile_x, y_prime - tile_y])

    coords_img = []
    for (x,y) in coords_tile:
      coords_img.append((int(x*tile_size), int(y*tile_size)))

    return coords_img

  

def splitZXY(zxy: str):
    """ Split an zxy string up in to z,x,y component

    Args:
        zxy (str): zxy string e.g. 12/345/678

    Returns:
        z: tile zoom level
        x: tile x
        y: tile y

    Example:
        Consider splitting the zxy string "12/345/678"::

            z, x, y = splitZXY("12/345/678")
            print(z,x,y)
            # result: 12, 345, 678

    """

    z = int(zxy.split("/")[0])
    x = int(zxy.split("/")[1])
    y = int(zxy.split("/")[2])

    return z, x, y


def urlFromZxy(z: int, x: int, y: int, imagery_id: str, baseUrl: str) -> str:
    """Generate a url given a zxy and an imagery id

    Args:
        z (int): zoom
        x (int): x tile
        y (int): y tile
        imagery_id (str): id of the imagery
        baseUrl (str): base url e.g. https://project-kiwi.org/

    Returns:
        str: url to download the tile, however api key is still required as a param
    """    
    return f"{baseUrl}/api/get_tile/{imagery_id}/{z}/{x}/{y}"


def maskFromPolygon(polygon: List[List], width: int, height: int) -> np.ndarray:
    """generates a binary mask from a polygon in image coordinates

    Args:
        polygon (List[List]): list of points in the polygon, image coordinates
        width (int): width of the output mask
        height (int): height of the output mask

    Returns:
        np.ndarray: the binary image
    """    
    im = Image.new('L', (width, height), 0)
    ImageDraw.Draw(im).polygon(polygon, outline=1, fill=1)
    mask = np.array(im)
    return mask


def bbox_iou(box1: List, box2: List) -> float:
    """calculate the intersection over union of two bounding boxes

    Args:
        box1 (List): first box
        box2 (List): second box

    Returns:
        float: the iou
    """    
    return iou(    [
                    [box1[0], box1[1]], 
                    [box1[2], box1[1]], 
                    [box1[2], box1[3]], 
                    [box1[0], box1[3]]
                    ],
                   [
                    [box2[0], box2[1]], 
                    [box2[2], box2[1]], 
                    [box2[2], box2[3]], 
                    [box2[0], box2[3]]
                    ])


def iou(poly1: List[List], poly2: List[List]) -> float:
    """ calculate the iou of each polygon

    Args:
        poly1 (List[List]): list of points in the first polygon e.g. [[x,y], [x,y]]
        poly2 (List[List]): list of points in the second polygon e.g. [[x,y], [x,y]]

    Returns:
        float: iou between the polygons.
    """
    
    poly1 = Polygon(poly1)
    poly2 = Polygon(poly2)
    intersection = poly1.intersection(poly2).area
    if intersection == 0:
      return 0
    union = poly1.union(poly2).area
    return intersection / union


from projectkiwi.tools import (
        bboxFromCoords,
        getAnnotationsForTile,
        latLngToImgCoords,
        maskFromPolygon,
        bbox_iou)

from PIL import Image
from pathlib import Path
import numpy as np
import numpy as np
import torch
from PIL import Image
import torch
from PIL import Image
from typing import List



class ProjectKiwiDataSet(object):

    def imgToTensor(self, img):
        if img.shape[-1] == 2:
            img = np.dstack((img[:,:,0], img[:,:,0], img[:,:,0]))
        assert img.shape[-1] == 3, f"Image must have three channels. expected: [h, w, 3] got: {img.shape}"
        img = torch.as_tensor(np.array(img), dtype=torch.float32)/255

        return img.permute(2, 0, 1)

    def getTaskTile(self, task):
        imageFile = str(self.cache_location / f"{self.imagery_id}" / f"padding_{self.padding}" / Path(task.zxy + ".png"))
        if not Path(imageFile).exists():
            # we can request a "super tile", which covers the same area but is at a higher resolution than a standard tile
            tile = self.conn.getSuperTile(self.imagery_id, task.zxy, self.max_zoom, self.padding)

            if tile.shape[2] == 4:
                tile = tile[:,:,:3]
            im = Image.fromarray(tile)
            Path(imageFile).parent.mkdir(parents=True, exist_ok=True)
            im.save(imageFile)
        else:
            im = Image.open(imageFile)
            tile = np.array(im)
        return tile

    def getAnnotations(self, project_id):
        annotationsAndPredictions = self.conn.getAnnotations(project_id)

        self.label_ids = []
        self.label_names = []
        for annotation in annotationsAndPredictions:
            if not annotation.label_id in self.label_ids:
                self.label_names.append(annotation.label_name)
                self.label_ids.append(annotation.label_id)

        # filter out predictions, we dont want to train on these
        self.annotations = [annotation for annotation in annotationsAndPredictions \
                    if annotation.confidence is None and annotation.shape == "Polygon"]


    def __init__(self, conn, tasks, project_id, imagery_id, max_zoom, 
            cache_location=Path("./cache"),
            padding = 0,
            inference = False,
            make_masks = True,
            transforms = None):
        self.conn = conn
        self.tasks = tasks
        self.imagery_id = imagery_id
        self.max_zoom = max_zoom
        self.cache_location = cache_location
        cache_location.mkdir(exist_ok=True)
        self.padding = padding
        self.inference = inference
        self.getAnnotations(project_id)
        self.make_masks = make_masks
        self.transforms = transforms
        


    def __getitem__(self, idx):
        task = self.tasks[idx]
        tile = self.getTaskTile(task)
        
        img = self.imgToTensor(tile)
        
        if self.inference == True:
            target = None
        else:
            annotationsForTile = getAnnotationsForTile(self.annotations, task.zxy, overlap_threshold=0.9)

            labels = []
            boxes = []
            isCrowd = []
            if self.make_masks:
                masks = np.zeros((len(annotationsForTile), tile.shape[0], tile.shape[1]))
            for i,annotation in enumerate(annotationsForTile):
                boxes.append(bboxFromCoords(annotation.coordinates, task.zxy, tile.shape[0], clip=False))

                # reserve label 0 for background
                labels.append(self.label_names.index(annotation.label_name) + 1)
                isCrowd.append(0)
                
                if self.make_masks:
                    poly = latLngToImgCoords(annotation.coordinates, task.zxy, tile.shape[0])
                    poly = [(x, y) for x,y in poly]
                    mask = maskFromPolygon(poly, tile.shape[0], tile.shape[1])
                    masks[i,:,:] = mask
               

            boxes = torch.as_tensor(boxes, dtype=torch.float32)
            labels = torch.as_tensor(labels, dtype=torch.int64)
            image_id = torch.tensor([idx])

            # empty targets for tiles without annotations
            if len(boxes) == 0:
                area = torch.zeros((1,), dtype=torch.int64)
                boxes = torch.zeros((0, 4), dtype=torch.float32)
            else:
                area = (boxes[:, 3] - boxes[:, 1]) * (boxes[:, 2] - boxes[:, 0])
            
            isCrowd = torch.as_tensor(isCrowd, dtype=torch.int64)

            target = {
                "boxes": boxes,
                "labels": labels,
                "image_id": image_id,
                "area": area,
                "iscrowd": isCrowd
            }

            if self.make_masks:
                target['masks'] = torch.as_tensor(masks, dtype=torch.uint8)

            if self.transforms is not None:
                img, target = self.transforms(img, target)

        return img, target, task

    def __len__(self):
        return len(self.tasks)




def scoreThresholding(boxes: List, scores: List, class_ids: List, masks: List = None, threshold = 0.1):
    """ apply a score threshold a list of boxes etc

    Args:
        boxes (List): bounding boxes for the objects
        scores (List): scores between 0 and 1
        class_ids (List): class ids
        masks (List, optional): masks if there is a mask-rcnn output. Defaults to None.
        threshold (float, optional): the threshold to apply, objects with scores below this omitted. Defaults to 0.1.

    Returns:
        _type_: the boxes, scores, ids and optionally masks
    """    
    returnBoxes, returnScores, returnIds = ([], [], [])
    if masks is None:
        returnMasks = None
    else:
        returnMasks = []

    for i, (box, score, class_id) in enumerate(zip(boxes, scores, class_ids)):

        if score > threshold:
            returnBoxes.append(box)
            returnScores.append(score)
            returnIds.append(class_id)
            if not masks is None:
                returnMasks.append(masks[i])

    return returnBoxes, returnScores, returnIds, returnMasks



def boxSizeFiltering(boxes: List, scores: List, class_ids: List, masks: List = None, min_side_length = 5):
    """ apply a filter on the smallest allowable side length for bounding boxes

    Args:
        boxes (List): bounding boxes for the objects
        scores (List): scores between 0 and 1
        class_ids (List): class ids
        masks (List, optional): masks if there is a mask-rcnn output. Defaults to None.
        min_side_length (int, optional): minimum allowed side length in pixels. Defaults to 5.

    Returns:
        _type_: the boxes, scores, ids and optionally masks
    """    
    returnBoxes, returnScores, returnIds = ([], [], [])
    if masks is None:
        returnMasks = None
    else:
        returnMasks = []

    for i, (box, score, class_id) in enumerate(zip(boxes, scores, class_ids)):
        [x1, y1, x2, y2] = [int(box[0]), int(box[1]), int(box[2]), int(box[3])]

        if y2-y1 >= min_side_length and x2-x1 >= min_side_length:
            returnBoxes.append(box)
            returnScores.append(score)
            returnIds.append(class_id)
            if not masks is None:
                returnMasks.append(masks[i])
    return returnBoxes, returnScores, returnIds, returnMasks




def nonMaximumSuppression(boxes: List, scores: List, class_ids: List, masks: List = None,
    iou_threshold = 0.3):
    """ apply non-maximum suppression to the output of an object detector

    Args:
        boxes (List): bounding boxes for the objects
        scores (List): scores between 0 and 1
        class_ids (List): class ids
        masks (List, optional): masks if there is a mask-rcnn output. Defaults to None.
        iou_threshold (float, optional): maximum overlap permitted, objects with a greater overlap with a higher confidence box will be omitted. Defaults to 0.3.

    Returns:
        _type_: the boxes, scores, ids and optionally masks
    """   

    returnBoxes, returnScores, returnIds = ([], [], [])
    if masks is None:
        returnMasks = None
    else:
        returnMasks = []

    for i, (box, score, class_id) in enumerate(zip(boxes, scores, class_ids)):
        skip = False
        for j, (box2, score2) in enumerate(zip(boxes, scores)):
            if i != j and score2 > score:
                if bbox_iou(box, box2) > iou_threshold:
                    skip = True
        
        if not skip: 
            returnBoxes.append(box)
            returnScores.append(score)
            returnIds.append(class_id)
            if not masks is None:
                returnMasks.append(masks[i])

    return returnBoxes, returnScores, returnIds, returnMasks




from re import L
import requests
import json
import numpy as np
from PIL import Image
import io
from typing import List
from projectkiwi.tools import getOverlap, num2deg, splitZXY, urlFromZxy
from projectkiwi.models import Annotation, Project, ImageryLayer, TilePath, Task, Label
import threading
import queue




class Connector():
    def __init__(self, key, url="https://project-kiwi.org/"):
        """constructor

        Args:
            key (str): API key.
            url (str, optional): url for api, in case of multiple instances. Defaults to "https://project-kiwi.org/api/".
        """

        self.key = key
        self.url = url


    def getImagery(self, project_id: str) -> List[ImageryLayer]:
        """Get a list of imagery layers for a project

        Args:
            project (str): ID of the project to get all the imagery for.

        Returns:
            List[ImageryLayer]: list of imagery layers
        """        
        
        route = "api/get_imagery"
        params = {
            'key': self.key, 
            'project_id': project_id
        }

        r = requests.get(self.url + route, params=params)
        r.raise_for_status()
        imageryList = r.json()
        imagery = []
        for layer in imageryList:
            imagery.append(ImageryLayer(**layer))
        assert len(imageryList) == len(imagery), "Failed to parse imagery"
        return imagery


    
    def getTile(self,
            imagery_id: str,
            z: int,
            x: int,
            y: int,
            tile_size: int = 256,
            tile_buffer: int = 0
        ) -> np.ndarray:
        """Download a tile given the z,x,y and id

        Args:
            imagery_id (str): id of the imagery
            z (int): zoom
            x (int): x tile
            y (int): y tile
            tile_size (int): width or height of the square tile
            tile_buffer (int): number of pixels to read each side of the tile


        Returns:
            np.ndarray: numpy array of tile
        """

        url = urlFromZxy(z, x, y, imagery_id, self.url)

        params={
            'key': self.key,
            'tile_size': tile_size,
            'tile_buffer': tile_buffer
        }

        r = requests.get(url, params=params)
        r.raise_for_status()
        tileContent = r.content
        return np.asarray(Image.open(io.BytesIO(tileContent)))
        




    def getTileList(self, 
            imagery_id: str, 
            project_id: str,
            zoom: int) -> List[TilePath]:
        """Get a list of tiles for a given imagery id

        Args:
            imageryId (str): ID of the imagery to retrieve a list of tiles for
            zoom (int): Zoom level

        Returns:
            List[TilePath]: A list of tiles with zxy and url
        """
        route = "api/get_tile_list"
        params = {
            'key': self.key, 
            'imagery_id': imagery_id, 
            'project_id': project_id,
            'zoom': zoom}

        r = requests.get(self.url + route, params=params)
        r.raise_for_status()
        tileList = r.json()
        tiles = []
        for tile in tileList:
            tiles.append(TilePath(**tile))
        assert len(tiles) == len(tileList), "Failed to parse tiles"
        return tiles


    def getImageryStatus(self, imagery_id: str) -> str:
        """ Get the status of imagery

        Args:
            imagery_id (str): Imagery id

        Returns:
            str: status
        """        
        route = "api/get_imagery_status"
        params = {'key': self.key, 'imagery_id': imagery_id}

        r = requests.get(self.url + route, params=params)
        r.raise_for_status()
        return r.json()['status']


    def getProjects(self) -> List[Project]:
        """Get a list of projects for a user

        Returns:
            List[Projects]: projects
        """
        route = "api/get_projects" 
        params = {'key': self.key}

        r = requests.get(self.url + route, params=params)
        r.raise_for_status()

        try:
            projectList = r.json()
            assert len(projectList) > 0, "Error: No projects found"
            projects = []
            for proj in projectList:
                projects.append(Project(**proj))
            assert len(projectList) == len(projects), \
                    f"Error: Could not parse projects, {projectList}"
            return projects
        except Exception as e:
            print("Error: Could not get projects")
            raise e
        

    def addImagery(self, filename: str, name: str, project_id: str) -> str:
        """ Add imagery to project-kiwi.org

        Args:
            filename (str): Path to the file to be uploaded
            name (str): Name for the imagery
            project_id (str): Id of the project to add the layer to

        Returns:
            str: imagery id
        """       
        
        # get presigned upload url
        route = "api/get_imagery_upload_url"
        params = {
            'key': self.key, 
            'filename': filename, 
            'name': name,
            'project_id': project_id
        }
        r = requests.get(self.url + route, params=params)
        r.raise_for_status()
        jsonResponse = r.json()
        url = jsonResponse['url']
        
        # upload
        with open(filename, 'rb') as data:
            r = requests.put(url, data=data, headers={'Content-type': ''})
            r.raise_for_status()

        return jsonResponse['imagery_id']

    def getSuperTile(self,
                imagery_id: str,
                zxy: str,
                max_zoom: int = 22,
                padding: int = 0
        ) -> np.ndarray:
        """Get a tile as higher resolution, as specified by the max zoom.

        Args:
            imagery_id (str): The ID of the imagery
            zxy (str): zxy string to specify the tile e.g. 12/345/678
            max_zoom (int, optional): Maximum zoom. Defaults to 22.
            padding (int, optional): Number of pixels to read on each side of the tile. Defaults to 0.

        Returns:
            np.ndarray: Image data for the tile.
        """            

        z,x,y = splitZXY(zxy)
        tile_width = 2**(max_zoom - z)
        width = 256*tile_width

        return self.getTile(imagery_id, z, x, y, tile_size=width, tile_buffer=padding)



    
    def getAnnotations(self, project_id: str) -> List[Annotation]:
        """Get all annotations in a project
        Args:
            project_id (str): id for the project to get the predictions for

        Returns:
            List[Annotation]: annotations
        """

        route = "api/get_annotations"
        params = {
            'key': self.key,
            'project_id': project_id
        }

        r = requests.get(self.url + route, params=params)
        r.raise_for_status()

        try:
            annotations = []
            annotationsDict = r.json()
            for annotation_id, data in annotationsDict.items():
                annotations.append(Annotation.from_dict(data, annotation_id))
            assert len(annotationsDict) == len(annotations), "ERROR: could not parse annotations"
            return annotations

        except Exception as e:
            print("Error: Could not load annotations")
            raise e


    def getPredictions(self, project_id: str) -> List[Annotation]:
        """Get all predictions in a project
        Args:
            project_id (str): id for the project to get the predictions for
        Returns:
            List[Annotation]: predictions
        """
        
        annotations = self.getAnnotations(project_id=project_id)

        return [annotation for annotation in annotations if annotation.confidence != None]



    def getTasks(self, queue_id: int) -> List[Task]:
        """Get a list of tasks in a queue.

        Args:
            queue_id (int): The ID of the queue

        Returns:
            List[Task]: list of tasks
        """        
       
        route = "api/get_tasks"
        params = {'key': self.key, "queue_id": queue_id}

        r = requests.get(self.url + route, params=params)
        r.raise_for_status()
        data = r.json()
        assert data['success'] == True, "Failed to get tasks"
        
        taskList = data['task']
        tasks = []
        for task in taskList:
            tasks.append(Task(**task))
        
        assert len(tasks) == len(taskList), "Failed to parse tasks"

        return tasks
    
    def getTask(self, queue_id: int) -> List[Task]:
        """Get a random task for a queue.

        Args:
            queue_id (int): The ID of the queue

        Returns:
            Task: task
        """        
       
        route = "api/get_task"
        params = {'key': self.key, "queue_id": queue_id}

        r = requests.get(self.url + route, params=params)
        r.raise_for_status()
        data = r.json()
        assert data['success'] == True, "Failed to get tasks"
        
        task = data['task']

        return Task(**task)

    def getNextTask(self, queue_id: int) -> List[Task]:
        """Get a predictable next task for a queue.

        Args:
            queue_id (int): The ID of the queue

        Returns:
            Task: task
        """        
       
        route = "api/get_next_task"
        params = {'key': self.key, "queue_id": queue_id}

        r = requests.get(self.url + route, params=params)
        r.raise_for_status()
        data = r.json()
        assert data['success'] == True, "Failed to get tasks"
        
        task = data['task']

        return Task(**task)


    def addAnnotation(self, annotation: Annotation, project: str) -> int:
        """Add an annotation to a project

        Args:
            annotation (Annotation): the annotation to add (note that not everything is mandatory)
            project (str): project id

        Returns:
            int: annotation id if successful
        """        
        route = "api/add_annotation"
        headers = {'Content-type': 'application/json', 'Accept': 'text/plain'}

        annoDict = dict(annotation)
        annoDict['project_id'] = project
        annoDict['key'] = self.key
        r = requests.post(self.url + route, data=json.dumps(annoDict), headers=headers)
        r.raise_for_status()
        jsonResponse = r.json()
        return jsonResponse['annotation_id']
    

    def addPrediction(self, annotation: Annotation, project: str) -> int:
        """Add a prediction to a project

        Args:
            annotation (Annotation): an annotation object with a confidence value
            project (str): project id

        Returns:
            int: annotation id if successful
        """       

        assert not annotation.confidence is None, "No confidence for prediction"
        route = "api/add_prediction"
        headers = {'Content-type': 'application/json', 'Accept': 'text/plain'}

        annoDict = dict(annotation)
        annoDict['project_id'] = project
        annoDict['key'] = self.key
        r = requests.post(self.url + route, data=json.dumps(annoDict), headers=headers)
        r.raise_for_status()
        jsonResponse = r.json()
        return jsonResponse['annotation_id']

    def getImageryUrl(self, imagery_id: str, project_id: str) -> str:
        """Get the url for imagery from it's id

        Args:
            imagery_id (str): Id for the imagery
            project_id (str): Project to look in

        Returns:
            str: The url template
        """        
        imagery = self.getImagery(project_id)
        imagery_url = [image.url for image in imagery if image.id == imagery_id][0]
        return imagery_url
        
    def removeAllPredictions(self, project_id: str):
        """Remove all predictions in a project

        Args:
            project_id (str): project id
        """    

        route = "api/remove_all_predictions" 
        params = {'key': self.key, 'project_id': project_id}

        r = requests.delete(self.url + route, 
                headers={'Content-Type': 'application/json'},
                data=json.dumps(params))
        r.raise_for_status()

    def getLabels(self, project_id: str) -> List[Label]:
        """Get all labels in a project
        Args:
            project_id (str): id for the project to get the labels for

        Returns:
            List[Label]: labels
        """

        route = "api/get_labels"
        params = {
            'key': self.key,
            'project_id': project_id
        }

        r = requests.get(self.url + route, params=params)
        r.raise_for_status()

        try:
            labels = []
            labelsJSON = r.json()
            for label in labelsJSON:
                labels.append(Label(**label))
            assert len(labelsJSON) == len(labels), "ERROR: could not parse labels"
            return labels

        except Exception as e:
            print("Error: Could not load labels")
            raise e
    


    def addLabel(self, name: str, project_id: str, color: str = None) -> Label:
        """ add a label to the project

        Args:
            name (str): name of the label e.g. object class
            project_id (str): id of the project that the label will belong to
            color (str, optional): Color string for the label e.g. rgb(255, 0, 100), will be selected randomly if not supplied. Defaults to None.

        Returns:
            Label: the label including it's id
        """        

        if color is None:
            rgb = list(np.random.choice(range(256), size=3))
            min_rgb = np.array(rgb).min()
            max_rgb = np.array(rgb).max()
            for i,c in enumerate(rgb):
                if c == min_rgb:
                    rgb[i] = 0
                if c == max_rgb:
                    rgb[i] = 256
            color = f"rgb({rgb[0]}, {rgb[1]}, {rgb[2]})"


        labelDict = {}
        labelDict['key'] = self.key
        labelDict['project_id'] = project_id
        labelDict['name'] = name
        labelDict['status'] = 'active'
        labelDict['color'] = color

        route = "api/add_label"
        headers = {'Content-type': 'application/json', 'Accept': 'text/plain'}


        r = requests.post(self.url + route, data=json.dumps(labelDict), headers=headers)
        r.raise_for_status()
        jsonResponse = r.json()
        return Label(**jsonResponse)


# !pip install projectkiwi
import sys
sys.path.append('../../projectkiwi/')

import projectkiwi
from projectkiwi.tools import (
        coordsFromPolygon,
        bboxToPolygon,
        yx_to_xy)
from projectkiwi.data import ProjectKiwiDataSet, scoreThresholding, boxSizeFiltering, nonMaximumSuppression

from tqdm import tqdm
from pathlib import Path
import numpy as np
from random import shuffle
import numpy as np
import torch
import torchvision
from torchvision.models.detection.faster_rcnn import FastRCNNPredictor
from torchvision.models.detection.mask_rcnn import MaskRCNNPredictor
import sys
import torch
import torchvision.models.detection.mask_rcnn
import threading


from skimage import measure
from skimage.measure import approximate_polygon



class BaseDetector(object):

    def threadAddPrediction(self, prediction):
        self.conn.addPrediction(prediction, self.project_id)

    @staticmethod
    def collate_fn(batch):
        return tuple(zip(*batch))

    def get_model():
        raise NotImplementedError("Please use a child class of this one")
    
    def __init__(self, 
            conn, 
            project_id, 
            imagery_id, 
            max_zoom, 
            tile_padding, 
            cache_location, 
            batch_size = 4, 
            model_load_path = None, 
            device = None, 
            transforms = None):

        self.conn = conn
        self.project_id = project_id
        self.imagery_id = imagery_id
        self.max_zoom = max_zoom
        self.tile_padding = tile_padding
        self.model = None
        self.cache_location = Path(cache_location)
        self.batch_size = batch_size
        self.class_names = None
        self.label_ids = None
        self.transforms = transforms
        
        if device != None:
            self.device = torch.device(device)
        
        self.device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')
        print(f"Using device: {self.device}")

        # Number of epochs to wait before terminating training without val improvement
        self.patience = 5

        if model_load_path == None:
            self.model_save_path = Path(self.cache_location / f"{self.project_id}_{self.model_name}.kiwi")
        else:
            self.model_save_path = model_load_path
        
        if Path(self.model_save_path).exists():
            state = torch.load(self.model_save_path)
            print(f"Loading model from: {self.model_save_path}")
            self.model = self.get_model(state['num_classes'])
            self.model.load_state_dict(state['state_dict'])
            self.class_names = state['class_names']
        else:
            print(f"No model found at: {self.model_save_path}")




    def train(self, tasks, max_epochs: int = 100, resume: bool = True) -> Path:
        print(f"Training for up to {max_epochs} epochs.")
        shuffle(tasks)

        assert len(tasks) > 0, "Please complete at least one task before training"

        num_examples = len(tasks)
        train_size = round(num_examples*0.8)
        test_size = num_examples - train_size

        dataset_train = ProjectKiwiDataSet(
                self.conn,
                tasks[:train_size],
                self.project_id,
                self.imagery_id,
                self.max_zoom,
                self.cache_location,
                make_masks=self.masks_required,
                transforms = self.transforms)

        data_loader_train = torch.utils.data.DataLoader(
                dataset_train,
                batch_size=self.batch_size,
                shuffle=True,
                num_workers=4,
                collate_fn=self.collate_fn)

        if test_size > 0:
            dataset_test = ProjectKiwiDataSet(
                    self.conn,
                    tasks[train_size:],
                    self.project_id,
                    self.imagery_id,
                    self.max_zoom,
                    self.cache_location,
                    make_masks=self.masks_required)
            data_loader_test = torch.utils.data.DataLoader(
                dataset_test,
                batch_size=self.batch_size,
                shuffle=True,
                num_workers=4,
                collate_fn=self.collate_fn)

        self.class_names = dataset_train.label_names
        self.label_ids = dataset_train.label_ids

        if self.model == None:
            self.model = self.get_model(num_classes = len(dataset_train.label_names)+1)
            if resume:
                if Path(self.model_save_path).exists():
                    state = torch.load(self.model_save_path)
                    print(f"Loading model from: {self.model_save_path}")
                    self.model.load_state_dict(state['state_dict'])
                else:
                    print(f"No model found at: {self.model_save_path}")
        self.model.to(self.device)

        # construct an optimizer
        params = [p for p in self.model.parameters() if p.requires_grad]
        optimizer = torch.optim.SGD(params, lr=0.001, momentum=0.9, weight_decay=0.0005)

        val_history = []


        for epoch in range(max_epochs):
            
            self.model.train()
            train_losses = []
            for images, targets, _ in data_loader_train:
                images = list(image.to(self.device) for image in images)
                targets = [{k: v.to(self.device) for k, v in t.items()} for t in targets]
                loss_dict = self.model(images, targets)
                losses = sum(loss for loss in loss_dict.values())
                train_losses.append(losses.item())
                
                optimizer.zero_grad()
                losses.backward()
                optimizer.step()

            train_loss = np.mean(train_losses)

            
            val_losses = []
            if test_size > 0:
                for images, targets, _ in data_loader_test:
                    images = list(image.to(self.device) for image in images)
                    targets = [{k: v.to(self.device) for k, v in t.items()} for t in targets]
                    loss_dict = self.model(images, targets)
                    val_losses.append(sum(loss for loss in loss_dict.values()).item())
                
                val_loss = np.mean(val_losses)
                
                val_history.append(val_loss)
                if len(val_history) >= self.patience:
                    if np.min(val_history[-5:]) > np.min(val_history):
                        print("stagnation detected, ending training!")
                        break

                print(f"Epoch {epoch+1}  Train loss: {train_loss:2.3f}  Val loss: {val_loss:2.3f}")
        
        print(f"Saving model to: {self.model_save_path}")
        state = {
            'num_classes': len(dataset_train.label_names)+1,
            'state_dict': self.model.state_dict(),
            'class_names': self.class_names
        }
        torch.save(state, self.model_save_path)

        return self.model_save_path

    
    def predict(self, tasks, remove_preds: bool = True):

        
        if remove_preds:
            print("Removing all predictions from project.")
            self.conn.removeAllPredictions(self.project_id)

        if self.label_ids is None:
            self.label_ids = []

            # get/make label ids for labels the model was trained on.
            labels = self.conn.getLabels(self.project_id)
            for class_name in self.class_names:
                label_id = None
                for label in labels:
                    if label.name == class_name:
                        label_id = label.id
                        break
                if label_id is None:
                    print(f"Creating label for: {class_name}")
                    label = self.conn.addLabel(
                            project_id = self.project_id,
                            name = class_name)
                    self.label_ids.append(label.id) 
                else:
                    self.label_ids.append(label.id)


        dataset = ProjectKiwiDataSet(self.conn, tasks, self.project_id, self.imagery_id, self.max_zoom, self.cache_location, self.tile_padding, inference=True, make_masks=False)
        data_loader_inference = torch.utils.data.DataLoader(dataset, batch_size=self.batch_size, shuffle=False, num_workers=4, collate_fn=self.collate_fn)

        self.model.to(self.device)
        self.model.eval()


        for images, _, tasks in tqdm(data_loader_inference, desc="Doing inference"):
            images = list(image.to(self.device) for image in images)
            results = self.model(images)
            for result, task in zip(results, tasks):

                tile_size = 256*2**(self.max_zoom - int(projectkiwi.tools.splitZXY(task.zxy)[0]))
                class_ids = result['labels'].detach().cpu().numpy()
                boxes = result['boxes'].detach().cpu().numpy()
                scores = result['scores'].detach().cpu().numpy()
                if self.masks_required:
                    masks = result['masks'].detach().cpu().numpy()
                else:
                    masks = None
                boxes, scores, class_ids, masks = scoreThresholding(boxes, scores, class_ids, masks)
                boxes, scores, class_ids, masks = boxSizeFiltering(boxes, scores, class_ids, masks)
                boxes, scores, class_ids, masks = nonMaximumSuppression(boxes, scores, class_ids, masks)
                
                if self.masks_required:
                    for box, score, class_id, mask in zip(boxes, scores, class_ids, masks):

                        contours = measure.find_contours(mask[0,:,:], 0.5)
                        if len(contours) == 0:
                            continue
                        contour = contours[0].astype(int)
                        approx_contour = yx_to_xy(approximate_polygon(contour, tolerance=5))

                        if len(approx_contour) < 4:
                            continue
                        
                        poly_latlng = coordsFromPolygon(approx_contour, task.zxy, tile_size, self.tile_padding)

                        
                        if poly_latlng[0] != poly_latlng[-1]:
                            continue

                        prediction = projectkiwi.models.Annotation(
                            shape="Polygon",
                            label_id=self.label_ids[class_id-1],
                            imagery_id=self.imagery_id,
                            coordinates=poly_latlng,
                            confidence = score)

                        threading.Thread(target=self.threadAddPrediction, args=(prediction,)).start()
                else:
                    for box, score, class_id in zip(boxes, scores, class_ids):
                        x1, y1, x2, y2 = [int(box[0]), int(box[1]), int(box[2]), int(box[3])]

                        latLngPoly = coordsFromPolygon(bboxToPolygon(x1, y1, x2, y2), task.zxy, tile_size, self.tile_padding)

                        # get the prediction ready
                        prediction = projectkiwi.models.Annotation(
                            shape="Polygon",
                            label_id=self.label_ids[class_id-1],
                            imagery_id=self.imagery_id,
                            coordinates=latLngPoly,
                            confidence = score)

                        # add the prediction (in a new thread for speeed)
                        threading.Thread(target=self.threadAddPrediction, args=(prediction,)).start()



class InstanceSegmentationModel(BaseDetector):
    masks_required = True
    model_name = "instance_seg"

    def get_model(self, num_classes):
        # load an instance segmentation model pre-trained on COCO
        model = torchvision.models.detection.maskrcnn_resnet50_fpn(weights="DEFAULT", trainable_backbone_layers=5)

        # get number of input features for the classifier
        in_features = model.roi_heads.box_predictor.cls_score.in_features

        # replace the pre-trained head with a new one
        model.roi_heads.box_predictor = FastRCNNPredictor(in_features, num_classes)

        # now get the number of input features for the mask classifier
        in_features_mask = model.roi_heads.mask_predictor.conv5_mask.in_channels
        hidden_layer = 256

        # and replace the mask predictor with a new one
        model.roi_heads.mask_predictor = MaskRCNNPredictor(in_features_mask, hidden_layer, num_classes)

        return model

class ObjectDetectionModel(BaseDetector):
    masks_required = False
    model_name = "obj_detector"

    def get_model(self, num_classes):
        # load an instance segmentation model pre-trained pre-trained on COCO
        model = torchvision.models.detection.fasterrcnn_resnet50_fpn(weights="DEFAULT")

        # get number of input features for the classifier
        in_features = model.roi_heads.box_predictor.cls_score.in_features
        
        # replace the pre-trained head with a new one
        model.roi_heads.box_predictor = FastRCNNPredictor(in_features, num_classes)

        return model
